/*
 * fftunzip.c -- Compressed file decompressor for Final Fantasy Tactics by
 *	Squaresoft.  This is not intended for copyright infringement.
 *
 * Copyright 2000-2002, George Greer <greerga@fftactics.org>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

/*
 * Patent commentary:
 *
 *  This decompresses the files marked ".lzw" in the "event/" directory
 *  of the Final Fantasy Tactics CD.  Based on the name, you'd think it
 *  was Lempel-Ziv-Welch compression, a modified form of Lempel-Ziv using
 *  a table.  Unisys holds a patent on that particular algorithm and is
 *  not very friendly about it: see the GIF issue. Curiously enough, this
 *  code to decompress doesn't use a table.  So are they really LZW files?
 *  No, likely not.  They match the description of LZ77-class compressors
 *  which use <offset,length> instead of the LZ78-class compressors (which
 *  LZW is one of) using dictionary offsets.
 *
 *    --George
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include "mmaplib.h"

static const char *program = __FILE__ ": Final Fantasy Tactics text file decompressor.";
static const char *version = "v1.0 (Build: "__DATE__" "__TIME__")";
static const char *copyright = "Copyright 2000-2002, George Greer <greerga@fftactics.org>";
static const char *license = "Licensed under GNU General Public License (GPL) v2 or later.";

#define FALSE	(0)
#define TRUE	(!FALSE)

struct options_s {
  unsigned char verbose;
};

/* ------------------------------------------------------------------------------------------------------------------------------ */

int memcspn(const void *haystack, const char *needles, size_t len)
{
  const char *hay = haystack;
  int it, pos = len;

  for (it = 0; it < strlen(needles); it++) {
    char *tops = memchr(hay, needles[it], len);
    if (tops && (tops - hay) < pos)
      pos = tops - hay;
  }
  return pos;
}

char tokens[] = { 0xf0, 0xf1, 0xf2, 0xf3, 0x0 };

int level = 0;
struct options_s options;

void resolve(const unsigned char *input, int rpos, unsigned char byte1, unsigned char byte2, unsigned char byte3)
{
  char translated[256];
  int pos, marker, txtlen, offset, junk;

  /* I don't think recursion is necessary now...? */
  if (level > 1000) {
    fputs("Run-away nested decompression.\n", stderr);
    exit(1);
  }

  /*
   * M = Marker.  / L = Length of text to take.
   * X = Unused.  / O = Offset back into the file.
   *
   * |---------- 1 ----------|   |---------- 2 ----------|   |---------- 3 ----------|
   * M3-M2-M1-M0   X?-X?-L4-L3   L2-L1-L0-OC   OB-OA-O9-O8   O7-O6-O5-O4   O3-O2-O1-O0
   */
  marker =  ((byte1 & 0xF0) >> 4);
  junk   =  ((byte1 & 0x0C) >> 2);
  txtlen = (((byte1 & 0x03) << 3) | ((byte2 & 0xE0) >> 5)) + 4;
  offset = (((byte2 & 0x1F) << 8) | byte3);

  if (options.verbose > 1)
    fprintf(stderr, "%6d: %2d: %#04hhx %#04hhx %#04hhx junk=%1d/%#4x len=%2d/%#4x offset=%4d/%#6x | %#04x%02x%02x = %x %x %x %x %x %x\n",
	rpos, level, byte1, byte2, byte3, junk, junk, txtlen, txtlen, offset, offset,
	input[rpos], input[rpos + 1], input[rpos + 2],
	(byte1 & 0xF0) >> 4, byte1 & 0x0F, (byte2 & 0xF0) >> 4, byte2 & 0x0F, (byte3 & 0xF0) >> 4, byte3 & 0x0F);

  /*
   * I'd really like an explanation of why this is required...
   */

  /* START KLUDGE */
  if (offset & (1 << 8))
    offset -= 2;
  if (offset & (1 << 9))
    offset -= 4;
  if (offset & (1 << 10))
    offset -= 8;
  if (offset & (1 << 11))
    offset -= 16;
  if (offset & (1 << 12))
    offset -= 32;
  /* END KLUDGE */

  if (offset == 0)
    return;

  memcpy(translated, input + rpos - offset, txtlen);
  pos = memcspn(translated, tokens, txtlen);
  if (pos == txtlen) {
    if (options.verbose) {
      putchar(0x8D);	/* ( */
      putchar(0xB5);	/* * */
    }

    fwrite(translated, sizeof(char), pos, stdout);

    if (options.verbose) {
      putchar(0xB5);	/* * */
      putchar(0x8E);	/* ) */
    }

    return;
  }

  /* Nested compression tokens if we get here. */

  if (options.verbose) {
    putchar(0x8D);	/* ( */
    putchar(0x8D);	/* ( */
  }

  if (pos > 0)
    fwrite(translated, sizeof(char), pos, stdout);

  if (options.verbose) {
    int i;

    for (i = 0; i < txtlen; i++)
      fprintf(stderr, "%4x%4x|", (translated[i] & 0xF0) >> 4, translated[i] & 0x0F);
    fputc('\n', stderr);
    fprintf(stderr, "pos: %d txtlen: %d\n", pos, txtlen);

    /* 0x3E = !   0x40 = ? */
    printf("%c%c%c%c%c%c%c", 0x3E, 0x40, translated[pos], translated[pos + 1], translated[pos + 2], 0x40, 0x3E);
  }

  /* FIXME: 'test.evt' causes runaway compression level. */
  if (txtlen - pos >= 3) {
    level++;
    resolve(input, rpos - offset + pos, translated[pos], translated[pos + 1], translated[pos + 2]);
    level--;
  }

  if (txtlen - pos > 3)
    fwrite(translated + pos + 3, sizeof(char), txtlen - pos - 3, stdout);

  if (options.verbose) {
    putchar(0x8E);	/* ) */
    putchar(0x8E);	/* ) */
  }
}

int fftunzip(const unsigned char *input, int length)
{
  unsigned int rpos = 0, wpos = 0;

  while (rpos < length) {
    switch (input[rpos]) {
      case 0xF0:
      case 0xF1:
      case 0xF2:
      case 0xF3:
        if (wpos != rpos)
          fwrite(input + wpos, sizeof(char), rpos - wpos, stdout);

        resolve(input, rpos, input[rpos], input[rpos + 1], input[rpos + 2]);
        rpos += 3;
        wpos = rpos;
        break;

      default:
        rpos++;
        break;
     }
  }

  if (wpos != rpos)
    fwrite(input + wpos, sizeof(char), rpos - wpos, stdout);

  return rpos;
}

/* ------------------------------------------------------------------------------------------------------------------------------ */

struct option /* {
  const char *name;
  int has_arg;
  int *flag;
  int val;
} */ long_options[] = {
  { "help",     FALSE, NULL, 0 },
  { "version",  FALSE, NULL, 1 },
  { 0, 0, 0, 0 }
};

int print_help(char *progname)
{
  printf("%s [-r] [-h] [-v] filename\n"
	 "Use -V or --version for version and copyright information.\n"
	 "%s --help	This screen.\n",
	progname, progname);
  return EXIT_FAILURE;
}

int print_version(char *progname)
{
  printf("%s\n\t%s\n\t%s\n\n%s: version %s\n", program, copyright, license, progname, version);
  return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
  void *handle;
  char *filename;
  int gto;

  options.verbose = FALSE;

  while ((gto = getopt_long(argc, argv, "Vv", long_options, NULL)) >= 0) {
    switch (gto) {
    case 0:
      return print_help(argv[0]);
    case 1:
    case 'V':
      return print_version(argv[0]);
    case 'v':
      options.verbose++;
      break;

    case ':':
      fprintf(stderr, "Option -%c requires an extra argument.\n", optopt);
      return EXIT_FAILURE;
    case '?':
      fprintf(stderr, "Unknown option -%c.\n", optopt);
      return EXIT_FAILURE;
    default:
      fputs("Unknown error condition.\n", stderr);
      return EXIT_FAILURE;
    }
  }

  if (argc == optind)
    fputs("Program requires a filename to translate.\nTry '--help' for more information.\n", stderr);
  else {
    while (optind < argc) {
      filename = argv[optind];
      if (!(handle = mapopen(filename, O_RDONLY))) {
        perror(filename);
        optind++;
        continue;
      }
      fftunzip(mapdata(handle), maplength(handle));
      mapclose(handle);
      optind++;
    }
    return EXIT_SUCCESS;
  }
  return EXIT_FAILURE;
}
