/*
 * mmaplib.c: Generic routines for File->Char*.
 *
 * Copyright (C) 2000, George Greer <greerga@fftactics.org>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/errno.h>

#include "mmaplib.h"

static const char *program = __FILE__ ": mmap() wrapper.";
static const char *version = "v0.8 (Build: "__DATE__" "__TIME__")";
static const char *copyright = "Copyright 2000, George Greer <greerga@fftactics.org>";
static const char *license = "Licensed under GNU General Public License (GPL) v2 or later.";

/* ----- MMAP Routines -------------------------------------------------------------------------------- */

int mapversion(const char *progname)
{
  printf("%s\n\t%s\n\t%s\n\n%s: version %s\n", program, copyright, license, progname, version);
  return EXIT_SUCCESS;
}

struct mmap_s {
  int fd;
  size_t size;
  void *map;
};

void *mapopen(const char *filename, int flags)
{
  int fd, mmapopt;
  struct stat filestat;
  struct mmap_s *ptrret;
  char *map;

  if ((fd = open(filename, flags)) < 0)
    error(filename);

  if (fstat(fd, &filestat) < 0)
    error("fstat");

  if (filestat.st_size == 0) {
    errno = ENODATA;
    return NULL;
  }

#ifdef WIN32
  map = (char *)calloc(1, filestat.st_size);
  read(fd, map, filestat.st_size);
#else
  mmapopt = 0;
  if ((flags & 3) == O_RDONLY) mmapopt = PROT_READ;
  if ((flags & 3) == O_WRONLY) mmapopt = PROT_WRITE;
  if ((flags & 3) == O_RDWR  ) mmapopt = PROT_READ | PROT_WRITE;
  if ((map = (char *)mmap(0, filestat.st_size, mmapopt, MAP_FILE | MAP_SHARED, fd, 0)) == MAP_FAILED)
    error("mmap");
#endif

  ptrret = (struct mmap_s *)calloc(1, sizeof(struct mmap_s));
  ptrret->fd = fd;
  ptrret->size = filestat.st_size;
  ptrret->map = map;

  return ptrret;
}

size_t maplength(void *handle)
{
  return ((struct mmap_s *)handle)->size;
}

char *mapdata(void *handle)
{
  return ((struct mmap_s *)handle)->map;
}

void mapclose(void *handle)
{
  struct mmap_s *localdata = (struct mmap_s *)handle;

#ifdef WIN32
  free(localdata->map);
#else
  if (munmap(localdata->map, localdata->size) < 0)
    error("munmap");
#endif

  if (close(localdata->fd) < 0)
    error("close");

  free(handle);
}

void error(const char *why)
{
  perror(why);
  exit(1);
}
